# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import unittest

from swh.model import exceptions, hashutil, validators


def hash_data(raw_content):
    return hashutil.MultiHash.from_data(raw_content).digest()


class TestValidators(unittest.TestCase):
    def setUp(self):
        self.valid_visible_content = {
            'status': 'visible',
            'length': 5,
            'data': b'1984\n',
            'ctime': datetime.datetime(2015, 11, 22, 16, 33, 56,
                                       tzinfo=datetime.timezone.utc),
        }

        self.valid_visible_content.update(
            hash_data(self.valid_visible_content['data']))

        self.valid_absent_content = {
            'status': 'absent',
            'length': 5,
            'ctime': datetime.datetime(2015, 11, 22, 16, 33, 56,
                                       tzinfo=datetime.timezone.utc),
            'reason': 'Content too large',
            'sha1_git': self.valid_visible_content['sha1_git'],
            'origin': 42,
        }

        self.invalid_content_hash_mismatch = self.valid_visible_content.copy()
        self.invalid_content_hash_mismatch.update(
            hash_data(b"this is not the data you're looking for"))

    def test_validate_content(self):
        self.assertTrue(
            validators.validate_content(self.valid_visible_content))

        self.assertTrue(
            validators.validate_content(self.valid_absent_content))

    def test_validate_content_hash_mismatch(self):
        with self.assertRaises(exceptions.ValidationError) as cm:
            validators.validate_content(self.invalid_content_hash_mismatch)

        # All the hashes are wrong. The exception should be of the form:
        # ValidationError({
        #     NON_FIELD_ERRORS: [
        #         ValidationError('content-hash-mismatch', 'sha1'),
        #         ValidationError('content-hash-mismatch', 'sha1_git'),
        #         ValidationError('content-hash-mismatch', 'sha256'),
        #     ]
        # })

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(set(exc.error_dict.keys()),
                         {exceptions.NON_FIELD_ERRORS})

        hash_mismatches = exc.error_dict[exceptions.NON_FIELD_ERRORS]
        self.assertIsInstance(hash_mismatches, list)
        self.assertEqual(len(hash_mismatches), 4)
        self.assertTrue(all(mismatch.code == 'content-hash-mismatch'
                            for mismatch in hash_mismatches))
        self.assertEqual(set(mismatch.params['hash']
                             for mismatch in hash_mismatches),
                         {'sha1', 'sha1_git', 'sha256', 'blake2s256'})



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import hashlib
import io
import os
import tempfile
import unittest
from unittest.mock import patch

from swh.model import hashutil
from swh.model.hashutil import MultiHash


class BaseHashutil(unittest.TestCase):
    def setUp(self):
        # Reset function cache
        hashutil._blake2_hash_cache = {}

        self.data = b'1984\n'
        self.hex_checksums = {
            'sha1': '62be35bf00ff0c624f4a621e2ea5595a049e0731',
            'sha1_git': '568aaf43d83b2c3df8067f3bedbb97d83260be6d',
            'sha256': '26602113b4b9afd9d55466b08580d3c2'
                      '4a9b50ee5b5866c0d91fab0e65907311',
            'blake2s256': '63cfb259e1fdb485bc5c55749697a6b21ef31fb7445f6c78a'
                          'c9422f9f2dc8906',
        }

        self.checksums = {
            type: bytes.fromhex(cksum)
            for type, cksum in self.hex_checksums.items()
        }

        self.bytehex_checksums = {
            type: hashutil.hash_to_bytehex(cksum)
            for type, cksum in self.checksums.items()
        }

        self.git_hex_checksums = {
            'blob': self.hex_checksums['sha1_git'],
            'tree': '5b2e883aa33d2efab98442693ea4dd5f1b8871b0',
            'commit': '79e4093542e72f0fcb7cbd75cb7d270f9254aa8f',
            'tag': 'd6bf62466f287b4d986c545890716ce058bddf67',
        }

        self.git_checksums = {
            type: bytes.fromhex(cksum)
            for type, cksum in self.git_hex_checksums.items()
        }


class MultiHashTest(BaseHashutil):
    def test_multi_hash_data(self):
        checksums = MultiHash.from_data(self.data).digest()
        self.assertEqual(checksums, self.checksums)
        self.assertFalse('length' in checksums)

    def test_multi_hash_data_with_length(self):
        expected_checksums = self.checksums.copy()
        expected_checksums['length'] = len(self.data)

        algos = set(['length']).union(hashutil.DEFAULT_ALGORITHMS)
        checksums = MultiHash.from_data(self.data, hash_names=algos).digest()

        self.assertEqual(checksums, expected_checksums)
        self.assertTrue('length' in checksums)

    def test_multi_hash_data_unknown_hash(self):
        with self.assertRaises(ValueError) as cm:
            MultiHash.from_data(self.data, ['unknown-hash'])

        self.assertIn('Unexpected hashing algorithm', cm.exception.args[0])
        self.assertIn('unknown-hash', cm.exception.args[0])

    def test_multi_hash_file(self):
        fobj = io.BytesIO(self.data)

        checksums = MultiHash.from_file(fobj, length=len(self.data)).digest()
        self.assertEqual(checksums, self.checksums)

    def test_multi_hash_file_hexdigest(self):
        fobj = io.BytesIO(self.data)
        length = len(self.data)
        checksums = MultiHash.from_file(fobj, length=length).hexdigest()
        self.assertEqual(checksums, self.hex_checksums)

    def test_multi_hash_file_bytehexdigest(self):
        fobj = io.BytesIO(self.data)
        length = len(self.data)
        checksums = MultiHash.from_file(fobj, length=length).bytehexdigest()
        self.assertEqual(checksums, self.bytehex_checksums)

    def test_multi_hash_file_missing_length(self):
        fobj = io.BytesIO(self.data)
        with self.assertRaises(ValueError) as cm:
            MultiHash.from_file(fobj, hash_names=['sha1_git'])

        self.assertIn('Missing length', cm.exception.args[0])

    def test_multi_hash_path(self):
        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.write(self.data)

        hashes = MultiHash.from_path(f.name).digest()
        os.remove(f.name)

        self.assertEqual(self.checksums, hashes)


class Hashutil(BaseHashutil):

    def test_hash_git_data(self):
        checksums = {
            git_type: hashutil.hash_git_data(self.data, git_type)
            for git_type in self.git_checksums
        }

        self.assertEqual(checksums, self.git_checksums)

    def test_hash_git_data_unknown_git_type(self):
        with self.assertRaises(ValueError) as cm:
            hashutil.hash_git_data(self.data, 'unknown-git-type')

        self.assertIn('Unexpected git object type', cm.exception.args[0])
        self.assertIn('unknown-git-type', cm.exception.args[0])

    def test_hash_to_hex(self):
        for type in self.checksums:
            hex = self.hex_checksums[type]
            hash = self.checksums[type]
            self.assertEqual(hashutil.hash_to_hex(hex), hex)
            self.assertEqual(hashutil.hash_to_hex(hash), hex)

    def test_hash_to_bytes(self):
        for type in self.checksums:
            hex = self.hex_checksums[type]
            hash = self.checksums[type]
            self.assertEqual(hashutil.hash_to_bytes(hex), hash)
            self.assertEqual(hashutil.hash_to_bytes(hash), hash)

    def test_hash_to_bytehex(self):
        for algo in self.checksums:
            self.assertEqual(self.hex_checksums[algo].encode('ascii'),
                             hashutil.hash_to_bytehex(self.checksums[algo]))

    def test_bytehex_to_hash(self):
        for algo in self.checksums:
            self.assertEqual(self.checksums[algo],
                             hashutil.bytehex_to_hash(
                                 self.hex_checksums[algo].encode()))

    def test_new_hash_unsupported_hashing_algorithm(self):
        try:
            hashutil._new_hash('blake2:10')
        except ValueError as e:
            self.assertEqual(str(e),
                             'Unexpected hashing algorithm blake2:10, '
                             'expected one of blake2b512, blake2s256, '
                             'sha1, sha1_git, sha256')

    @patch('hashlib.new')
    def test_new_hash_blake2b_blake2b512_builtin(self, mock_hashlib_new):
        if 'blake2b512' not in hashlib.algorithms_available:
            self.skipTest('blake2b512 not built-in')
        mock_hashlib_new.return_value = sentinel = object()

        h = hashutil._new_hash('blake2b512')

        self.assertIs(h, sentinel)
        mock_hashlib_new.assert_called_with('blake2b512')

    @patch('hashlib.new')
    def test_new_hash_blake2s_blake2s256_builtin(self, mock_hashlib_new):
        if 'blake2s256' not in hashlib.algorithms_available:
            self.skipTest('blake2s256 not built-in')
        mock_hashlib_new.return_value = sentinel = object()

        h = hashutil._new_hash('blake2s256')

        self.assertIs(h, sentinel)
        mock_hashlib_new.assert_called_with('blake2s256')

    def test_new_hash_blake2b_builtin(self):
        removed_hash = False

        try:
            if 'blake2b512' in hashlib.algorithms_available:
                removed_hash = True
                hashlib.algorithms_available.remove('blake2b512')
            if 'blake2b' not in hashlib.algorithms_available:
                self.skipTest('blake2b not built in')

            with patch('hashlib.blake2b') as mock_blake2b:
                mock_blake2b.return_value = sentinel = object()

                h = hashutil._new_hash('blake2b512')

                self.assertIs(h, sentinel)
                mock_blake2b.assert_called_with(digest_size=512//8)
        finally:
            if removed_hash:
                hashlib.algorithms_available.add('blake2b512')

    def test_new_hash_blake2s_builtin(self):
        removed_hash = False

        try:
            if 'blake2s256' in hashlib.algorithms_available:
                removed_hash = True
                hashlib.algorithms_available.remove('blake2s256')
            if 'blake2s' not in hashlib.algorithms_available:
                self.skipTest('blake2s not built in')

            with patch('hashlib.blake2s') as mock_blake2s:
                mock_blake2s.return_value = sentinel = object()

                h = hashutil._new_hash('blake2s256')

                self.assertIs(h, sentinel)
                mock_blake2s.assert_called_with(digest_size=256//8)
        finally:
            if removed_hash:
                hashlib.algorithms_available.add('blake2s256')

    def test_new_hash_blake2b_pyblake2(self):
        if 'blake2b512' in hashlib.algorithms_available:
            self.skipTest('blake2b512 built in')
        if 'blake2b' in hashlib.algorithms_available:
            self.skipTest('blake2b built in')

        with patch('pyblake2.blake2b') as mock_blake2b:
            mock_blake2b.return_value = sentinel = object()

            h = hashutil._new_hash('blake2b512')

            self.assertIs(h, sentinel)
            mock_blake2b.assert_called_with(digest_size=512//8)

    def test_new_hash_blake2s_pyblake2(self):
        if 'blake2s256' in hashlib.algorithms_available:
            self.skipTest('blake2s256 built in')
        if 'blake2s' in hashlib.algorithms_available:
            self.skipTest('blake2s built in')

        with patch('pyblake2.blake2s') as mock_blake2s:
            mock_blake2s.return_value = sentinel = object()

            h = hashutil._new_hash('blake2s256')

            self.assertIs(h, sentinel)
            mock_blake2s.assert_called_with(digest_size=256//8)


class HashlibGit(unittest.TestCase):

    def setUp(self):
        self.blob_data = b'42\n'

        self.tree_data = b''.join([b'40000 barfoo\0',
                                   bytes.fromhex('c3020f6bf135a38c6df'
                                                 '3afeb5fb38232c5e07087'),
                                   b'100644 blah\0',
                                   bytes.fromhex('63756ef0df5e4f10b6efa'
                                                 '33cfe5c758749615f20'),
                                   b'100644 hello\0',
                                   bytes.fromhex('907b308167f0880fb2a'
                                                 '5c0e1614bb0c7620f9dc3')])

        self.commit_data = """tree 1c61f7259dcb770f46b194d941df4f08ff0a3970
author Antoine R. Dumont (@ardumont) <antoine.romain.dumont@gmail.com> 1444054085 +0200
committer Antoine R. Dumont (@ardumont) <antoine.romain.dumont@gmail.com> 1444054085 +0200

initial
""".encode('utf-8')  # NOQA
        self.tag_data = """object 24d012aaec0bc5a4d2f62c56399053d6cc72a241
type commit
tag 0.0.1
tagger Antoine R. Dumont (@ardumont) <antoine.romain.dumont@gmail.com> 1444225145 +0200

blah
""".encode('utf-8')  # NOQA

        self.checksums = {
            'blob_sha1_git': bytes.fromhex('d81cc0710eb6cf9efd5b920a8453e1'
                                           'e07157b6cd'),
            'tree_sha1_git': bytes.fromhex('ac212302c45eada382b27bfda795db'
                                           '121dacdb1c'),
            'commit_sha1_git': bytes.fromhex('e960570b2e6e2798fa4cfb9af2c399'
                                             'd629189653'),
            'tag_sha1_git': bytes.fromhex('bc2b99ba469987bcf1272c189ed534'
                                          'e9e959f120'),
        }

    def test_unknown_header_type(self):
        with self.assertRaises(ValueError) as cm:
            hashutil.hash_git_data(b'any-data', 'some-unknown-type')

        self.assertIn('Unexpected git object type', cm.exception.args[0])

    def test_hashdata_content(self):
        # when
        actual_hash = hashutil.hash_git_data(self.blob_data, git_type='blob')

        # then
        self.assertEqual(actual_hash,
                         self.checksums['blob_sha1_git'])

    def test_hashdata_tree(self):
        # when
        actual_hash = hashutil.hash_git_data(self.tree_data, git_type='tree')

        # then
        self.assertEqual(actual_hash,
                         self.checksums['tree_sha1_git'])

    def test_hashdata_revision(self):
        # when
        actual_hash = hashutil.hash_git_data(self.commit_data,
                                             git_type='commit')

        # then
        self.assertEqual(actual_hash,
                         self.checksums['commit_sha1_git'])

    def test_hashdata_tag(self):
        # when
        actual_hash = hashutil.hash_git_data(self.tag_data, git_type='tag')

        # then
        self.assertEqual(actual_hash,
                         self.checksums['tag_sha1_git'])



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import attr
from hypothesis import given

from swh.model.hashutil import DEFAULT_ALGORITHMS
from swh.model.hypothesis_strategies import objects, object_dicts


target_types = (
    'content', 'directory', 'revision', 'release', 'snapshot', 'alias')


@given(objects())
def test_generation(obj_type_and_obj):
    (obj_type, object_) = obj_type_and_obj
    attr.validate(object_)


def assert_nested_dict(obj):
    """Tests the object is a nested dict and contains no more class
    from swh.model.model."""
    if isinstance(obj, dict):
        for (key, value) in obj.items():
            assert isinstance(key, (str, bytes)), key
            assert_nested_dict(value)
    elif isinstance(obj, list):
        for value in obj:
            assert_nested_dict(value)
    elif isinstance(obj, (int, float, str, bytes, bool, type(None))):
        pass
    else:
        assert False, obj


@given(object_dicts())
def test_dicts_generation(obj_type_and_obj):
    (obj_type, object_) = obj_type_and_obj
    assert_nested_dict(object_)
    if obj_type == 'content':
        if object_['status'] == 'visible':
            assert set(object_) == \
                set(DEFAULT_ALGORITHMS) | {'length', 'status', 'data'}
        elif object_['status'] == 'absent':
            assert set(object_) == \
                set(DEFAULT_ALGORITHMS) | {'length', 'status', 'reason'}
        elif object_['status'] == 'hidden':
            assert set(object_) == \
                set(DEFAULT_ALGORITHMS) | {'length', 'status'}
        else:
            assert False, object_
    elif obj_type == 'release':
        assert object_['target_type'] in target_types
    elif obj_type == 'snapshot':
        for branch in object_['branches'].values():
            assert branch['target_type'] in target_types



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import defaultdict
import itertools

from ..exceptions import ValidationError, NON_FIELD_ERRORS


def validate_against_schema(model, schema, value):
    """Validate a value for the given model against the given schema.

    Args:
        model: the name of the model
        schema: the schema to validate against
        value: the value to validate

    Returns:
        True if the value is correct against the schema

    Raises:
        ValidationError if the value does not validate against the schema
    """

    if not isinstance(value, dict):
        raise ValidationError(
            'Unexpected type %(type)s for %(model)s, expected dict',
            params={
                'model': model,
                'type': value.__class__.__name__,
            },
            code='model-unexpected-type',
        )

    errors = defaultdict(list)

    for key, (mandatory, validators) in itertools.chain(
        ((k, v) for k, v in schema.items() if k != NON_FIELD_ERRORS),
        [(NON_FIELD_ERRORS, (False, schema.get(NON_FIELD_ERRORS, [])))]
    ):
        if not validators:
            continue

        if not isinstance(validators, list):
            validators = [validators]

        validated_value = value
        if key != NON_FIELD_ERRORS:
            try:
                validated_value = value[key]
            except KeyError:
                if mandatory:
                    errors[key].append(
                        ValidationError(
                            'Field %(field)s is mandatory',
                            params={'field': key},
                            code='model-field-mandatory',
                        )
                    )

                continue
        else:
            if errors:
                # Don't validate the whole object if some fields are broken
                continue

        for validator in validators:
            try:
                valid = validator(validated_value)
            except ValidationError as e:
                errors[key].append(e)
            else:
                if not valid:
                    errdata = {
                        'validator': validator.__name__,
                    }

                    if key == NON_FIELD_ERRORS:
                        errmsg = 'Validation of model %(model)s failed in ' \
                                 '%(validator)s'
                        errdata['model'] = model
                        errcode = 'model-validation-failed'
                    else:
                        errmsg = 'Validation of field %(field)s failed in ' \
                                 '%(validator)s'
                        errdata['field'] = key
                        errcode = 'field-validation-failed'

                    errors[key].append(
                        ValidationError(errmsg, params=errdata, code=errcode)
                    )

    if errors:
        raise ValidationError(dict(errors))

    return True


def validate_all_keys(value, keys):
    """Validate that all the given keys are present in value"""
    missing_keys = set(keys) - set(value)
    if missing_keys:
        missing_fields = ', '.join(sorted(missing_keys))
        raise ValidationError(
            'Missing mandatory fields %(missing_fields)s',
            params={'missing_fields': missing_fields},
            code='missing-mandatory-field'
        )

    return True


def validate_any_key(value, keys):
    """Validate that any of the given keys is present in value"""
    present_keys = set(keys) & set(value)
    if not present_keys:
        missing_fields = ', '.join(sorted(keys))
        raise ValidationError(
            'Must contain one of the alternative fields %(missing_fields)s',
            params={'missing_fields': missing_fields},
            code='missing-alternative-field',
        )

    return True



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from swh.model.exceptions import ValidationError
from swh.model.fields import hashes


class ValidateHashes(unittest.TestCase):
    def setUp(self):
        self.valid_byte_hashes = {
            'sha1': b'\xf1\xd2\xd2\xf9\x24\xe9\x86\xac\x86\xfd\xf7\xb3\x6c\x94'
                    b'\xbc\xdf\x32\xbe\xec\x15',
            'sha1_git': b'\x25\x7c\xc5\x64\x2c\xb1\xa0\x54\xf0\x8c\xc8\x3f\x2d'
                        b'\x94\x3e\x56\xfd\x3e\xbe\x99',
            'sha256': b'\xb5\xbb\x9d\x80\x14\xa0\xf9\xb1\xd6\x1e\x21\xe7\x96'
                      b'\xd7\x8d\xcc\xdf\x13\x52\xf2\x3c\xd3\x28\x12\xf4\x85'
                      b'\x0b\x87\x8a\xe4\x94\x4c',
        }

        self.valid_str_hashes = {
            'sha1': 'f1d2d2f924e986ac86fdf7b36c94bcdf32beec15',
            'sha1_git': '257cc5642cb1a054f08cc83f2d943e56fd3ebe99',
            'sha256': 'b5bb9d8014a0f9b1d61e21e796d78dccdf1352f23cd32812f485'
                      '0b878ae4944c',
        }

        self.bad_hash = object()

    def test_valid_bytes_hash(self):
        for hash_type, value in self.valid_byte_hashes.items():
            self.assertTrue(hashes.validate_hash(value, hash_type))

    def test_valid_str_hash(self):
        for hash_type, value in self.valid_str_hashes.items():
            self.assertTrue(hashes.validate_hash(value, hash_type))

    def test_invalid_hash_type(self):
        hash_type = 'unknown_hash_type'
        with self.assertRaises(ValidationError) as cm:
            hashes.validate_hash(self.valid_str_hashes['sha1'], hash_type)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, 'unexpected-hash-type')
        self.assertEqual(exc.params['hash_type'], hash_type)

        self.assertIn('Unexpected hash type', str(exc))
        self.assertIn(hash_type, str(exc))

    def test_invalid_bytes_len(self):
        for hash_type, value in self.valid_byte_hashes.items():
            value = value + b'\x00\x01'
            with self.assertRaises(ValidationError) as cm:
                hashes.validate_hash(value, hash_type)

            exc = cm.exception
            self.assertIsInstance(str(exc), str)
            self.assertEqual(exc.code, 'unexpected-hash-length')
            self.assertEqual(exc.params['hash_type'], hash_type)
            self.assertEqual(exc.params['length'], len(value))

            self.assertIn('Unexpected length', str(exc))
            self.assertIn(str(len(value)), str(exc))

    def test_invalid_str_len(self):
        for hash_type, value in self.valid_str_hashes.items():
            value = value + '0001'
            with self.assertRaises(ValidationError) as cm:
                hashes.validate_hash(value, hash_type)

            exc = cm.exception
            self.assertIsInstance(str(exc), str)
            self.assertEqual(exc.code, 'unexpected-hash-length')
            self.assertEqual(exc.params['hash_type'], hash_type)
            self.assertEqual(exc.params['length'], len(value))

            self.assertIn('Unexpected length', str(exc))
            self.assertIn(str(len(value)), str(exc))

    def test_invalid_str_contents(self):
        for hash_type, value in self.valid_str_hashes.items():
            value = '\xa2' + value[1:-1] + '\xc3'
            with self.assertRaises(ValidationError) as cm:
                hashes.validate_hash(value, hash_type)

            exc = cm.exception
            self.assertIsInstance(str(exc), str)
            self.assertEqual(exc.code, 'unexpected-hash-contents')
            self.assertEqual(exc.params['hash_type'], hash_type)
            self.assertEqual(exc.params['unexpected_chars'], '\xa2, \xc3')

            self.assertIn('Unexpected characters', str(exc))
            self.assertIn('\xc3', str(exc))
            self.assertIn('\xa2', str(exc))

    def test_invalid_value_type(self):
        with self.assertRaises(ValidationError) as cm:
            hashes.validate_hash(self.bad_hash, 'sha1')

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, 'unexpected-hash-value-type')
        self.assertEqual(exc.params['type'], self.bad_hash.__class__.__name__)

        self.assertIn('Unexpected type', str(exc))
        self.assertIn(self.bad_hash.__class__.__name__, str(exc))

    def test_validate_sha1(self):
        self.assertTrue(hashes.validate_sha1(self.valid_byte_hashes['sha1']))
        self.assertTrue(hashes.validate_sha1(self.valid_str_hashes['sha1']))

        with self.assertRaises(ValidationError) as cm:
            hashes.validate_sha1(self.bad_hash)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, 'unexpected-hash-value-type')
        self.assertEqual(exc.params['type'], self.bad_hash.__class__.__name__)

    def test_validate_sha1_git(self):
        self.assertTrue(
            hashes.validate_sha1_git(self.valid_byte_hashes['sha1_git']))
        self.assertTrue(
            hashes.validate_sha1_git(self.valid_str_hashes['sha1_git']))

        with self.assertRaises(ValidationError) as cm:
            hashes.validate_sha1_git(self.bad_hash)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, 'unexpected-hash-value-type')
        self.assertEqual(exc.params['type'], self.bad_hash.__class__.__name__)

    def test_validate_sha256(self):
        self.assertTrue(
            hashes.validate_sha256(self.valid_byte_hashes['sha256']))
        self.assertTrue(
            hashes.validate_sha256(self.valid_str_hashes['sha256']))

        with self.assertRaises(ValidationError) as cm:
            hashes.validate_sha256(self.bad_hash)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, 'unexpected-hash-value-type')
        self.assertEqual(exc.params['type'], self.bad_hash.__class__.__name__)



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

from hypothesis.strategies import (
    binary, builds, characters, composite, dictionaries, from_regex,
    integers, just, lists, none, one_of, sampled_from, text, tuples,
)


from .from_disk import DentryPerms
from .model import (
    Person, Timestamp, TimestampWithTimezone, Origin, OriginVisit,
    Snapshot, SnapshotBranch, TargetType, Release, Revision,
    Directory, DirectoryEntry, Content
)
from .identifiers import snapshot_identifier, identifier_to_bytes


def sha1_git():
    return binary(min_size=20, max_size=20)


def sha1():
    return binary(min_size=20, max_size=20)


@composite
def urls(draw):
    protocol = draw(sampled_from(['git', 'http', 'https', 'deb']))
    domain = draw(from_regex(r'\A([a-z]([a-z0-9-]*)\.){1,3}[a-z0-9]+\Z'))

    return '%s://%s' % (protocol, domain)


def persons():
    return builds(Person)


def timestamps():
    max_seconds = datetime.datetime.max.replace(
        tzinfo=datetime.timezone.utc).timestamp()
    min_seconds = datetime.datetime.min.replace(
        tzinfo=datetime.timezone.utc).timestamp()
    return builds(
        Timestamp,
        seconds=integers(min_seconds, max_seconds),
        microseconds=integers(0, 1000000))


def timestamps_with_timezone():
    return builds(
        TimestampWithTimezone,
        timestamp=timestamps(),
        offset=integers(min_value=-14*60, max_value=14*60))


def origins():
    return builds(
        Origin,
        type=sampled_from(['git', 'hg', 'svn', 'pypi', 'deb']),
        url=urls())


def origin_visits():
    return builds(
        OriginVisit,
        visit=integers(0, 1000),
        origin=origins())


@composite
def releases(draw):
    (date, author) = draw(one_of(
        tuples(none(), none()),
        tuples(timestamps_with_timezone(), persons())))
    rel = draw(builds(
        Release,
        id=sha1_git(),
        author=none(),
        date=none(),
        target=sha1_git()))
    rel.date = date
    rel.author = author
    return rel


def revision_metadata():
    alphabet = characters(
        blacklist_categories=('Cs', ),
        blacklist_characters=['\u0000'])  # postgresql does not like these
    return dictionaries(text(alphabet=alphabet), text(alphabet=alphabet))


def revisions():
    return builds(
        Revision,
        id=sha1_git(),
        date=timestamps_with_timezone(),
        committer_date=timestamps_with_timezone(),
        parents=lists(sha1_git()),
        directory=sha1_git(),
        metadata=one_of(none(), revision_metadata()))
    # TODO: metadata['extra_headers'] can have binary keys and values


def directory_entries():
    return builds(
        DirectoryEntry,
        target=sha1_git(),
        perms=sampled_from([perm.value for perm in DentryPerms]))


def directories():
    return builds(
        Directory,
        id=sha1_git(),
        entries=lists(directory_entries()))


@composite
def contents(draw):
    (status, data, reason) = draw(one_of(
        tuples(just('visible'), binary(), none()),
        tuples(just('absent'), none(), text()),
        tuples(just('hidden'), none(), none()),
    ))

    return draw(builds(
        Content,
        length=integers(min_value=0, max_value=2**63-1),
        sha1=sha1(),
        sha1_git=sha1_git(),
        sha256=binary(min_size=32, max_size=32),
        blake2s256=binary(min_size=32, max_size=32),
        status=just(status),
        data=just(data),
        reason=just(reason),
    ))


def branch_names():
    return binary()


def branch_targets_object():
    return builds(
        SnapshotBranch,
        target=sha1_git(),
        target_type=sampled_from([
            TargetType.CONTENT, TargetType.DIRECTORY, TargetType.REVISION,
            TargetType.RELEASE, TargetType.SNAPSHOT]))


def branch_targets_alias():
    return builds(
        SnapshotBranch,
        target_type=just(TargetType.ALIAS))


def branch_targets(*, only_objects=False):
    if only_objects:
        return branch_targets_object()
    else:
        return one_of(branch_targets_alias(), branch_targets_object())


@composite
def snapshots(draw, *, min_size=0, max_size=100, only_objects=False):
    branches = draw(dictionaries(
        keys=branch_names(),
        values=branch_targets(only_objects=only_objects),
        min_size=min_size,
        max_size=max_size,
    ))

    if not only_objects:
        # Make sure aliases point to actual branches
        unresolved_aliases = {
            target.target
            for target in branches.values()
            if (target
                and target.target_type == 'alias'
                and target.target not in branches)
         }

        for alias in unresolved_aliases:
            branches[alias] = draw(branch_targets(only_objects=True))

    while True:
        try:
            id_ = snapshot_identifier({
                'branches': {
                    name: branch.to_dict()
                    for (name, branch) in branches.items()}})
        except ValueError as e:
            for (source, target) in e.args[1]:
                branches[source] = draw(branch_targets(only_objects=True))
        else:
            break
    return Snapshot(
        id=identifier_to_bytes(id_),
        branches=branches)


def objects():
    return one_of(
        origins().map(lambda x: ('origin', x)),
        origin_visits().map(lambda x: ('origin_visit', x)),
        snapshots().map(lambda x: ('snapshot', x)),
        releases().map(lambda x: ('release', x)),
        revisions().map(lambda x: ('revision', x)),
        directories().map(lambda x: ('directory', x)),
        contents().map(lambda x: ('content', x)),
    )


def object_dicts():
    return objects().map(lambda x: (x[0], x[1].to_dict()))



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import click
import os
import sys

from functools import partial

from swh.model import identifiers as pids
from swh.model.exceptions import ValidationError
from swh.model.from_disk import Content, Directory


class PidParamType(click.ParamType):
    name = 'persistent identifier'

    def convert(self, value, param, ctx):
        try:
            pids.parse_persistent_identifier(value)
            return value  # return as string, as we need just that
        except ValidationError as e:
            self.fail('%s is not a valid PID. %s.' % (value, e), param, ctx)


def pid_of_file(path):
    object = Content.from_file(path=path).get_data()
    return pids.persistent_identifier(pids.CONTENT, object)


def pid_of_dir(path):
    object = Directory.from_disk(path=path).get_data()
    return pids.persistent_identifier(pids.DIRECTORY, object)


def identify_object(obj_type, follow_symlinks, obj):
    if obj_type == 'auto':
        if os.path.isfile(obj):
            obj_type = 'content'
        elif os.path.isdir(obj):
            obj_type = 'directory'
        else:  # shouldn't happen, due to path validation
            raise click.BadParameter('%s is neither a file nor a directory' %
                                     obj)

    path = obj
    if follow_symlinks and os.path.islink(obj):
        path = os.path.realpath(obj)

    pid = None
    if obj_type == 'content':
        pid = pid_of_file(path)
    elif obj_type == 'directory':
        pid = pid_of_dir(path)
    else:  # shouldn't happen, due to option validation
        raise click.BadParameter('invalid object type: ' + obj_type)

    # note: we return original obj instead of path here, to preserve user-given
    # file name in output
    return (obj, pid)


@click.command()
@click.option('--dereference/--no-dereference', 'follow_symlinks',
              default=True,
              help='follow (or not) symlinks for OBJECTS passed as arguments '
              + '(default: follow)')
@click.option('--filename/--no-filename', 'show_filename', default=True,
              help='show/hide file name (default: show)')
@click.option('--type', '-t', 'obj_type', default='auto',
              type=click.Choice(['auto', 'content', 'directory']),
              help='type of object to identify (default: auto)')
@click.option('--verify', '-v', metavar='PID', type=PidParamType(),
              help='reference identifier to be compared with computed one')
@click.argument('objects', nargs=-1,
                type=click.Path(exists=True, readable=True,
                                allow_dash=True, path_type=bytes))
def identify(obj_type, verify, show_filename, follow_symlinks, objects):
    """Compute the Software Heritage persistent identifier (PID) for the given
    source code object(s).

    For more details about Software Heritage PIDs see:

    \b
    https://docs.softwareheritage.org/devel/swh-model/persistent-identifiers.html

    \b
    Examples:

    \b
      $ swh-identify fork.c kmod.c sched/deadline.c
      swh:1:cnt:2e391c754ae730bd2d8520c2ab497c403220c6e3    fork.c
      swh:1:cnt:0277d1216f80ae1adeed84a686ed34c9b2931fc2    kmod.c
      swh:1:cnt:57b939c81bce5d06fa587df8915f05affbe22b82    sched/deadline.c

    \b
      $ swh-identify --no-filename /usr/src/linux/kernel/
      swh:1:dir:f9f858a48d663b3809c9e2f336412717496202ab

    """
    if verify and len(objects) != 1:
        raise click.BadParameter('verification requires a single object')

    results = map(partial(identify_object, obj_type, follow_symlinks), objects)

    if verify:
        pid = next(results)[1]
        if verify == pid:
            click.echo('PID match: %s' % pid)
            sys.exit(0)
        else:
            click.echo('PID mismatch: %s != %s' % (verify, pid))
            sys.exit(1)
    else:
        for (obj, pid) in results:
            msg = pid
            if show_filename:
                msg = '%s\t%s' % (pid, os.fsdecode(obj))
            click.echo(msg)


if __name__ == '__main__':
    identify()



# Copyright (C) 2018-2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from enum import Enum
from typing import List, Optional, Dict

import attr


# TODO: Limit this to 20 bytes
Sha1Git = bytes


@attr.s
class Person:
    name = attr.ib(type=bytes)
    email = attr.ib(type=bytes)
    fullname = attr.ib(type=bytes)


@attr.s
class Timestamp:
    seconds = attr.ib(type=int)
    microseconds = attr.ib(type=int)

    @seconds.validator
    def check_seconds(self, attribute, value):
        """Check that seconds fit in a 64-bits signed integer."""
        if not (-2**63 <= value < 2**63):
            raise ValueError('Seconds must be a signed 64-bits integer.')

    @microseconds.validator
    def check_microseconds(self, attribute, value):
        """Checks that microseconds are positive and < 1000000."""
        if not (0 <= value < 10**6):
            raise ValueError('Microseconds must be in [0, 1000000[.')


@attr.s
class TimestampWithTimezone:
    timestamp = attr.ib(type=Timestamp)
    offset = attr.ib(type=int)
    negative_utc = attr.ib(type=bool)

    def to_dict(self):
        return attr.asdict(self)

    @offset.validator
    def check_offset(self, attribute, value):
        if not (-2**15 <= value < 2**15):
            # max 14 hours offset in theory, but you never know what
            # you'll find in the wild...
            raise ValueError('offset too large: %d minutes' % value)


@attr.s
class Origin:
    type = attr.ib(type=str)
    url = attr.ib(type=str)

    def to_dict(self):
        return attr.asdict(self)


@attr.s
class OriginVisit:
    origin = attr.ib(type=Origin)
    date = attr.ib(type=datetime.datetime)
    visit = attr.ib(type=Optional[int])
    """Should not be set before calling 'origin_visit_add()'."""

    def to_dict(self):
        ov = attr.asdict(self)
        ov['origin'] = self.origin.to_dict()
        ov['date'] = str(self.date)
        if not ov['visit']:
            del ov['visit']
        return ov


class TargetType(Enum):
    CONTENT = 'content'
    DIRECTORY = 'directory'
    REVISION = 'revision'
    RELEASE = 'release'
    SNAPSHOT = 'snapshot'
    ALIAS = 'alias'


class ObjectType(Enum):
    CONTENT = 'content'
    DIRECTORY = 'directory'
    REVISION = 'revision'
    RELEASE = 'release'
    SNAPSHOT = 'snapshot'


@attr.s
class SnapshotBranch:
    target = attr.ib(type=bytes)
    target_type = attr.ib(type=TargetType)

    @target.validator
    def check_target(self, attribute, value):
        if self.target_type != TargetType.ALIAS:
            if len(value) != 20:
                raise ValueError('Wrong length for bytes identifier: %d' %
                                 len(value))

    def to_dict(self):
        branch = attr.asdict(self)
        branch['target_type'] = branch['target_type'].value
        return branch


@attr.s
class Snapshot:
    id = attr.ib(type=Sha1Git)
    branches = attr.ib(type=Dict[bytes, Optional[SnapshotBranch]])

    def to_dict(self):
        return {
            'id': self.id,
            'branches': {
                name: branch.to_dict()
                for (name, branch) in self.branches.items()
            }
        }


@attr.s
class Release:
    id = attr.ib(type=Sha1Git)
    name = attr.ib(type=bytes)
    message = attr.ib(type=bytes)
    date = attr.ib(type=Optional[TimestampWithTimezone])
    author = attr.ib(type=Optional[Person])
    target = attr.ib(type=Optional[Sha1Git])
    target_type = attr.ib(type=ObjectType)
    synthetic = attr.ib(type=bool)

    def to_dict(self):
        rel = attr.asdict(self)
        rel['date'] = self.date.to_dict() if self.date is not None else None
        rel['target_type'] = rel['target_type'].value
        return rel

    @author.validator
    def check_author(self, attribute, value):
        if self.author is None and self.date is not None:
            raise ValueError('release date must be None if date is None.')


class RevisionType(Enum):
    GIT = 'git'
    TAR = 'tar'
    DSC = 'dsc'
    SUBVERSION = 'svn'
    MERCURIAL = 'hg'


@attr.s
class Revision:
    id = attr.ib(type=Sha1Git)
    message = attr.ib(type=bytes)
    author = attr.ib(type=Person)
    committer = attr.ib(type=Person)
    date = attr.ib(type=TimestampWithTimezone)
    committer_date = attr.ib(type=TimestampWithTimezone)
    parents = attr.ib(type=List[Sha1Git])
    type = attr.ib(type=RevisionType)
    directory = attr.ib(type=Sha1Git)
    metadata = attr.ib(type=Optional[Dict[str, object]])
    synthetic = attr.ib(type=bool)

    def to_dict(self):
        rev = attr.asdict(self)
        rev['date'] = self.date.to_dict()
        rev['committer_date'] = self.committer_date.to_dict()
        rev['type'] = rev['type'].value
        return rev


@attr.s
class DirectoryEntry:
    name = attr.ib(type=bytes)
    type = attr.ib(type=str,
                   validator=attr.validators.in_(['file', 'dir', 'rev']))
    target = attr.ib(type=Sha1Git)
    perms = attr.ib(type=int)
    """Usually one of the values of `swh.model.from_disk.DentryPerms`."""

    def to_dict(self):
        return attr.asdict(self)


@attr.s
class Directory:
    id = attr.ib(type=Sha1Git)
    entries = attr.ib(type=List[DirectoryEntry])

    def to_dict(self):
        dir_ = attr.asdict(self)
        dir_['entries'] = [entry.to_dict() for entry in self.entries]
        return dir_


@attr.s
class Content:
    sha1 = attr.ib(type=bytes)
    sha1_git = attr.ib(type=Sha1Git)
    sha256 = attr.ib(type=bytes)
    blake2s256 = attr.ib(type=bytes)

    data = attr.ib(type=bytes)
    length = attr.ib(type=int)
    status = attr.ib(
        type=str,
        validator=attr.validators.in_(['visible', 'absent', 'hidden']))
    reason = attr.ib(type=Optional[str])

    @length.validator
    def check_length(self, attribute, value):
        """Checks the length is positive."""
        if value < 0:
            raise ValueError('Length must be positive.')

    @reason.validator
    def check_reason(self, attribute, value):
        """Checks the reason is full iff status != absent."""
        assert self.reason == value
        if self.status == 'absent' and value is None:
            raise ValueError('Must provide a reason if content is absent.')
        elif self.status != 'absent' and value is not None:
            raise ValueError(
                'Must not provide a reason if content is not absent.')

    def to_dict(self):
        content = attr.asdict(self)
        if content['data'] is None:
            del content['data']
        if content['reason'] is None:
            del content['reason']
        return content
